@@ -47,15 +47,28 @@ class DataFrame:
             - a row is returned as a dictionary of 0d torch.Tensor, 0d np.ndarray, or object
             - a sub-frame is returned as a DataFrame
 
-        :param data: a 2D np.ndarray/torch.Tensor, a dictionary of (0/1D torch.Tensor/np.ndarray or any other object), or a DataFrame
+        :param data: a 2D np.ndarray/torch.Tensor, a list/dictionary of (0/1D torch.Tensor/np.ndarray or any other object), or a DataFrame
         :param columns: a list of column names (optional)
         """
         if columns is not None:
-            if not (isinstance(columns, collections.abc.Iterable) and isinstance(columns, collections.abc.Sized)) or isinstance(columns, np.ndarray) or isinstance(columns, torch.Tensor):
+            if not ((isinstance(columns, collections.abc.Iterable) and isinstance(columns, collections.abc.Sized)) or isinstance(columns, np.ndarray) or isinstance(columns, torch.Tensor)):
                 raise TypeError(f"columns must be None or be a sized iterable, but {type(columns)=}.")
                 
             if len(set(columns)) != len(columns):
                 raise ValueError(f"Columns must be unique, but detected duplicates in {columns}.")
+                
+        if isinstance(data, list):
+            if len(data) > 0 and isinstance(data[0], dict):
+                data = {k: v for k, *v in lib.itertools.dict_zip(*data, mode='strict')}
+            
+            else:
+                _columns = range(len(data)) if columns is None else columns
+
+                if len(_columns) != len(data):
+                    raise ValueError(f"length of columns must be same as length of data, but {len(_columns)=} and {len(data)=}.")
+
+                data = {k: v for k, v in zip(_columns, data)}
+                columns = None
         
         if isinstance(data, dict):
             if columns is not None:
@@ -106,6 +119,35 @@ class DataFrame:
                 
         self._data = data
         
+    @classmethod
+    def empty(cls, N_rows, dtype=None, is_tensor=False, device=None, columns=None, masked=True):
+        if dtype is None:
+            dtype = float
+        if device is None:
+            device = 'cpu'
+        
+        dtype = np.atleast_1d(dtype)
+        is_tensor = np.atleast_1d(is_tensor)
+        device = np.atleast_1d(device)
+        
+        dtype, is_tensor, device = np.broadcast_arrays(dtype, is_tensor, device)
+        
+        if columns is not None and len(columns) != len(dtype):
+            raise ValueError("length of columns is incompatible with other provided arguments.")
+            
+        data = [ar.Array.empty((N_rows,), dtype=dt, is_tensor=it, device=dv, masked=masked) for dt, it, dv in zip(dtype, is_tensor, device)]
+        
+        return cls(data, columns=columns, copy=False)
+    
+    @classmethod
+    def empty_like(cls, df, N_rows=None, masked=True):
+        if N_rows is None:
+            N_rows = len(df)
+            
+        info = df.info()
+        
+        return cls.empty(N_rows, dtype=info['dtype'], is_tensor=info['is_tensor'], device=info['device'], columns=info['column'], masked=masked)
+        
     def _validate(self):
         lengths = {}
         for column, data in self._items():
@@ -303,6 +345,12 @@ class DataFrame:
         """
         return self.copy()._data
     
+    def as_dict(self):
+        """
+        Returns data as a dict of columns. Data is not copied (only a shallow dict copy is performed).
+        """
+        return self._data.copy()
+    
     def to_numpy(self):
         """
         Returns data as a 2D np.ndarray. Data is copied (but non-array objects are not copied).
@@ -396,7 +444,7 @@ class DataFrame:
         return arr
     
     def info(self):
-        return pd.DataFrame([{
+        return DataFrame([{
             'column': k,
             'n_rows': len(v), # should be same for every column
             'is_tensor': v.is_tensor,
@@ -440,10 +488,15 @@ class DataFrame:
         return self[:,self.where(condition, level=level+1)]
     
     def merge(self, df, how='inner', on=None, left_on=None, right_on=None, suffixes=('_x', '_y')):
+        """
+        Note on efficiency: For i such that left_on[i] != right_on[i], this function returns both left_on[i] and right_on[i]
+        resulting in increased memory usage. So for best performance, rename the columns you want to merge so that
+        left_on[i] == right_on[i] for all i before merging.
+        """
         # if not isinstance(df, DataFrame):
         #     raise TypeError(f"df must be a DataFrame, but {type(df)=}.")
         
-        if not how == 'inner':
+        if how == 'cross':
             raise NotImplementedError()
         
         if on is not None:
@@ -498,12 +551,26 @@ class DataFrame:
         # logger.debug(f'{lindices=}')
         # logger.debug(f'{rindices=}')
         
-        left, right = self[:,lindices], df[:,rindices]
-        del right[[column for column in left_on if column in right_on]]
+        on = [column for column in left_on if column in right_on]
+        left_not_on = [column for column in self.columns if column not in on]
+        right_not_on = [column for column in df.columns if column not in on]
+        
+        left, middle, right = self[left_not_on, lindices], self[on, lindices], df[right_not_on, rindices]
+        
+        if how in ['left', 'outer']:
+            left = concat([self[left_not_on, ~lmask], left])
+            middle = concat([self[on, ~lmask], middle])
+            right = concat([DataFrame.empty_like(right, N_rows=(~lmask).sum()), right])
+            
+        if how in ['right', 'outer']:
+            left = concat([left, DataFrame.empty_like(left, N_rows=(~rmask).sum())])
+            middle = concat([middle, df[on, ~rmask]])
+            right = concat([right, df[right_not_on, ~rmask]])
+            
         left_rename = {column: f'{column}{suffixes[0]}' for column in left.columns if column in right.columns}
         right_rename = {column: f'{column}{suffixes[1]}' for column in right.columns if column in left.columns}
-        
-        return concat([left.rename(left_rename), right.rename(right_rename)], axis=1)
+
+        return concat([left.rename(left_rename), middle, right.rename(right_rename)], axis=1)
     
 class DataFrameGroupBy:
     def __init__(self, df, by, sort=True, dropna=True):
@@ -536,39 +603,81 @@ class DataFrameGroupBy:
                 continue
                 
             yield tuple(group), self.df[columns, self._row_to_group_idx.data == idx]
+            
+    def __getitem__(self, key):
+        return ColumnGroupBy(self, key)
         
-    def _agg(self, column, func, results):
-        if (column, func) in results:
-            return results[(column, func)]
+    def agg(self, *args, **kwargs):
+        new_args = collections.defaultdict(list)
+        for arg in args:
+            if not (isinstance(arg, tuple) and len(arg) == 2):
+                raise TypeError(f"arg must be a 2-tuple of (column_name, func), but {arg=}.")
+                
+            column, func = arg
+            new_args[column].append(func)
         
-        _row_to_group_idx = self._row_to_group_idx.like(self.df[column], dtype=False)
+        new_kwargs = collections.defaultdict(dict)
+        for k, v in kwargs.items():
+            if not (isinstance(v, tuple) and len(v) == 2):
+                raise TypeError(f"v must be a 2-tuple of (column_name, func), but {v=}.")
+                
+            column, func = v
+            new_kwargs[column][k] = func
+        
+        results = {}
+        for column, args, kwargs in lib.itertools.dict_zip(new_args, new_kwargs, mode='union'):
+            args = [] if args is None else args
+            kwargs = {} if kwargs is None else kwargs
+            result = self[column].agg(*args, **kwargs)
+            if isinstance(result, DataFrame):
+                results = results | {k if k in kwargs else f'{column}_{k}': v for k, v in result.items()}
+            else:
+                results[column] = result
+            
+        return DataFrame(self.groups.as_dict() | results)
+    
+class ColumnGroupBy:
+    def __init__(self, dfgb, column):
+        self._row_to_group_idx = dfgb._row_to_group_idx
+        self._groups = dfgb._groups
+        self._na = dfgb._na
+        self.groups = dfgb.groups
+        self.dropna = dfgb.dropna
+        self.df = dfgb.df
+        self.column = column
+        
+    def _agg(self, func, results):
+        if func in results:
+            return results[func]
+        
+        _row_to_group_idx = self._row_to_group_idx.like(self.df[self.column], dtype=False)
 
         # See scipy.stats.binned_statistic_dd for reference
         if func == 'count':
             result = _row_to_group_idx.bincount()
         
         elif func == 'sum':
-            result = _row_to_group_idx.bincount(weights=self.df[column])
+            result = _row_to_group_idx.bincount(weights=self.df[self.column])
         
         elif func == 'mean':
-            result = self._agg(column, 'sum', results) / self._agg(column, 'count', results)
+            result = self._agg('sum', results) / self._agg('count', results)
             
         elif func == 'var':
-            count, mean = self._agg(column, 'count', results), self._agg(column, 'mean', results)
+            count, mean = self._agg('count', results), self._agg('mean', results)
             result = (_row_to_group_idx.bincount(
-                weights=(self.df[column] - mean[_row_to_group_idx].data)**2
+                weights=(self.df[self.column] - mean[_row_to_group_idx].data)**2
             ))/(count-1)
             
         elif func == 'std':
-            var = self._agg(column, 'var', results)
+            var = self._agg('var', results)
             result = var**0.5
         
         elif func == 'sem':
-            count, std = self._agg(column, 'count', results), self._agg(column, 'std', results)
+            count, std = self._agg('count', results), self._agg('std', results)
             result = std / count**0.5
         
         elif func == 'min':
-            v = ar.Array(self.df[column])
+            v = ar.Array(self.df[self.column])
             i = v.argsort(descending=True)
             result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
             
@@ -580,7 +689,7 @@ class DataFrameGroupBy:
                 torch.use_deterministic_algorithms(False)
         
         elif func == 'max':
-            v = ar.Array(self.df[column])
+            v = ar.Array(self.df[self.column])
             i = v.argsort()
             result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
             
@@ -592,7 +701,7 @@ class DataFrameGroupBy:
                 torch.use_deterministic_algorithms(False)
         
         elif isinstance(func, callable):
-            v = ar.Array(self.df[column])
+            v = ar.Array(self.df[self.column])
             result = ar.Array.empty((len(self._groups),), masked=False, is_tensor=v.is_tensor, dtype=v.dtype, device=v.device)
             for idx, group in enumerate(self._groups):
                 result[idx] = func(v[_row_to_group_idx == idx].data)
@@ -600,17 +709,26 @@ class DataFrameGroupBy:
         else:
             raise ValueError(f"func must be a callable or a string in ['count', 'sum', 'mean', 'var', 'std', 'sem', 'min', 'max'], but got {func}")
         return result
+    
+    def agg(self, *args, **kwargs):
+        N = len(args) + len(kwargs)
+        if N == 0:
+            raise ValueError("must provide arguments to agg.")
+            
+        funcs = {arg if isinstance(arg, str) else arg.__name__: arg for arg in args} | kwargs
         
-    def agg(self, **kwargs):
-        data = {k: v for k, v in self.groups.items()}
+        if len(funcs) != N:
+            raise ValueError("provided args with duplicate names, please provide unique names using keyword arguments instead.")
+            
         results = {}
-        for _, v in kwargs.items():
-            results[tuple(v)] = self._agg(*v, results)
+        for name, func in funcs.items():
+            results[name] = self._agg(func, results)
             
         results = {k: v[~self._na].data if self.dropna else v.data for k, v in results.items()}
-        data = data | {k: results[v] for k, v in kwargs.items()}
         
-        return DataFrame(data)
+        if len(results) == 1:
+            return next(iter(results.values()))
+        return DataFrame(results)
         
 def concat(dfs, axis=0):
     assert len(dfs) > 0
@@ -621,7 +739,7 @@ def concat(dfs, axis=0):
             raise ValueError(f"dfs must have the same columns when axis=0, but {all_columns=}.")
             
         columns = all_columns[0]
-        return DataFrame({column: ar.concat([df[column] for df in dfs]) for column in columns})
+        return DataFrame({column: ar.concat([df._data[column] for df in dfs]) for column in columns})
     
     elif axis == 1:
         columns = lib.itertools.flatten_seq(all_columns)
